from __future__ import annotations

import numpy as np
import tensorflow.keras.backend as kbackend
import tensorflow.keras.layers as klayers

import innvestigate.analyzer.relevance_based.utils as rutils
import innvestigate.backend as ibackend
import innvestigate.backend.graph as igraph
import innvestigate.layers as ilayers
from innvestigate.backend.types import Layer, OptionalList, Tensor

# TODO: differentiate between LRP and DTD rules?
# DTD rules are special cases of LRP rules with additional assumptions
__all__ = [
    # dedicated treatment for special layers
    # general rules
    "ZRule",
    "EpsilonRule",
    "WSquareRule",
    "FlatRule",
    "AlphaBetaRule",
    "AlphaBetaIgnoreBiasRule",
    "Alpha2Beta1Rule",
    "Alpha2Beta1IgnoreBiasRule",
    "Alpha1Beta0Rule",
    "Alpha1Beta0IgnoreBiasRule",
    "AlphaBetaXRule",
    "AlphaBetaX1000Rule",
    "AlphaBetaX1010Rule",
    "AlphaBetaX1001Rule",
    "AlphaBetaX2m100Rule",
    "ZPlusRule",
    "ZPlusFastRule",
    "BoundedRule",
]


class ZRule(igraph.ReverseMappingBase):
    """
    Basic LRP decomposition rule (for layers with weight kernels),
    which considers the bias a constant input neuron.
    """

    def __init__(self, layer: Layer, _state, bias: bool = True) -> None:
        self._layer_wo_act = igraph.copy_layer_wo_activation(
            layer, keep_bias=bias, name_template="reversed_kernel_%s"
        )

    def apply(
        self,
        Xs: list[Tensor],
        _Ys: list[Tensor],
        Rs: list[Tensor],
        _reverse_state,
    ) -> list[Tensor]:
        # Get activations.
        Zs = ibackend.apply(self._layer_wo_act, Xs)
        # Divide incoming relevance by the activations.
        tmp = [ibackend.safe_divide(a, b) for a, b in zip(Rs, Zs)]
        # Propagate the relevance to input neurons
        # using the gradient.
        grads = ibackend.gradients(Xs, Zs, tmp)
        # Re-weight relevance with the input values.
        return [klayers.Multiply()([a, b]) for a, b in zip(Xs, grads)]


class EpsilonRule(igraph.ReverseMappingBase):
    """
    Similar to ZRule.
    The only difference is the addition of a numerical stabilizer term
    epsilon to the decomposition function's denominator.
    the sign of epsilon depends on the sign of the output activation
    0 is considered to be positive, ie sign(0) = 1
    """

    def __init__(self, layer: Layer, _state, epsilon=1e-7, bias: bool = True):
        self._epsilon = rutils.assert_lrp_epsilon_param(epsilon, self)
        self._layer_wo_act = igraph.copy_layer_wo_activation(
            layer, keep_bias=bias, name_template="reversed_kernel_%s"
        )

    def apply(
        self,
        Xs: list[Tensor],
        _Ys: list[Tensor],
        Rs: list[Tensor],
        _reverse_state: dict,
    ):
        # The epsilon rule aligns epsilon with the (extended) sign:
        # 0 is considered to be positive
        prepare_div = klayers.Lambda(
            lambda x: x
            + (kbackend.cast(kbackend.greater_equal(x, 0), kbackend.floatx()) * 2 - 1)
            * self._epsilon
        )

        # Get activations.
        Zs = ibackend.apply(self._layer_wo_act, Xs)

        # Divide incoming relevance by the activations.
        tmp = [a / prepare_div(b) for a, b in zip(Rs, Zs)]
        # Propagate the relevance to input neurons
        # using the gradient.
        grads = ibackend.gradients(Xs, Zs, tmp)
        # Re-weight relevance with the input values.
        return [klayers.Multiply()([a, b]) for a, b in zip(Xs, grads)]


class WSquareRule(igraph.ReverseMappingBase):
    """W**2 rule from Deep Taylor Decomposition"""

    def __init__(self, layer: Layer, _state, copy_weights=False) -> None:
        # W-square rule works with squared weights and no biases.
        if copy_weights:
            weights = layer.get_weights()
        else:
            weights = layer.weights
        if getattr(layer, "use_bias", False):
            weights = weights[:-1]
        weights = [x**2 for x in weights]

        self._layer_wo_act_b = igraph.copy_layer_wo_activation(
            layer, keep_bias=False, weights=weights, name_template="reversed_kernel_%s"
        )

    def apply(
        self,
        Xs: list[Tensor],
        Ys: list[Tensor],
        Rs: list[Tensor],
        _reverse_state: dict,
    ) -> list[Tensor]:
        # Create dummy forward path to take the derivative below.
        Ys = ibackend.apply(self._layer_wo_act_b, Xs)

        # Compute the sum of the weights.
        ones = ilayers.OnesLike()(Xs)
        Zs = [self._layer_wo_act_b(X) for X in ones]
        # Weight the incoming relevance.
        tmp = [ilayers.SafeDivide()([a, b]) for a, b in zip(Rs, Zs)]
        # Redistribute the relevances along the gradient.
        grads = ibackend.gradients(Xs, Ys, tmp)
        return grads


class FlatRule(WSquareRule):
    """Same as W**2 rule but sets all weights to ones."""

    def __init__(self, layer: Layer, _state, copy_weights: bool = False) -> None:
        # The flat rule works with weights equal to one and
        # no biases.
        if copy_weights:
            weights = layer.get_weights()
            if getattr(layer, "use_bias", False):
                weights = weights[:-1]
            weights = [np.ones_like(x) for x in weights]
        else:
            weights = layer.weights
            if getattr(layer, "use_bias", False):
                weights = weights[:-1]
            weights = [kbackend.ones_like(x) for x in weights]

        self._layer_wo_act_b = igraph.copy_layer_wo_activation(
            layer, keep_bias=False, weights=weights, name_template="reversed_kernel_%s"
        )


class AlphaBetaRule(igraph.ReverseMappingBase):
    """
    This decomposition rule handles the positive forward
    activations (x*w > 0) and negative forward activations
    (w * x < 0) independently, reducing the risk of zero
    divisions considerably. In fact, the only case where
    divisions by zero can happen is if there are either
    no positive or no negative parts to the activation
    at all.
    Corresponding parameterization of this rule implement
    methods such as Excitation Backpropagation with
    alpha=1, beta=0
    s.t.
    alpha - beta = 1 (after current param. scheme.)
    and
    alpha > 1
    beta > 0
    """

    def __init__(
        self,
        layer: Layer,
        _state,
        alpha=None,
        beta=None,
        bias: bool = True,
        copy_weights=False,
    ) -> None:
        alpha, beta = rutils.assert_infer_lrp_alpha_beta_param(alpha, beta, self)
        self._alpha = alpha
        self._beta = beta

        # prepare positive and negative weights for computing positive
        # and negative preactivations z in apply_accordingly.
        if copy_weights:
            weights = layer.get_weights()
            if not bias and getattr(layer, "use_bias", False):
                weights = weights[:-1]
            positive_weights = [x * (x > 0) for x in weights]
            negative_weights = [x * (x < 0) for x in weights]
        else:
            weights = layer.weights
            if not bias and getattr(layer, "use_bias", False):
                weights = weights[:-1]
            positive_weights = [x * ibackend.cast_to_floatx(x > 0) for x in weights]
            negative_weights = [x * ibackend.cast_to_floatx(x < 0) for x in weights]

        self._layer_wo_act_positive = igraph.copy_layer_wo_activation(
            layer,
            keep_bias=bias,
            weights=positive_weights,
            name_template="reversed_kernel_positive_%s",
        )
        self._layer_wo_act_negative = igraph.copy_layer_wo_activation(
            layer,
            keep_bias=bias,
            weights=negative_weights,
            name_template="reversed_kernel_negative_%s",
        )

    def apply(
        self,
        Xs: list[Tensor],
        _Ys: list[Tensor],
        Rs: list[Tensor],
        _reverse_state: dict,
    ):
        # this method is correct, but wasteful
        times_alpha = klayers.Lambda(lambda x: x * self._alpha)
        times_beta = klayers.Lambda(lambda x: x * self._beta)
        keep_positives = klayers.Lambda(
            lambda x: x * kbackend.cast(kbackend.greater(x, 0), kbackend.floatx())
        )
        keep_negatives = klayers.Lambda(
            lambda x: x * kbackend.cast(kbackend.less(x, 0), kbackend.floatx())
        )

        def fn_tmp(
            layer1: Layer,
            layer2: Layer,
            X1: OptionalList[Tensor],
            X2: OptionalList[Tensor],
        ):
            # Get activations of full positive or negative part.
            Z1 = ibackend.apply(layer1, X1)
            Z2 = ibackend.apply(layer2, X2)
            Zs = [klayers.Add()([a, b]) for a, b in zip(Z1, Z2)]
            # Divide incoming relevance by the activations.
            tmp = [ibackend.safe_divide(a, b) for a, b in zip(Rs, Zs)]
            # Propagate the relevance to the input neurons
            # using the gradient
            grads1 = ibackend.gradients(X1, Z1, tmp)
            grads2 = ibackend.gradients(X2, Z2, tmp)
            # Re-weight relevance with the input values.
            tmp1 = [klayers.Multiply()([a, b]) for a, b in zip(X1, grads1)]
            tmp2 = [klayers.Multiply()([a, b]) for a, b in zip(X2, grads2)]
            # combine and return
            return [klayers.Add()([a, b]) for a, b in zip(tmp1, tmp2)]

        # Distinguish postive and negative inputs.
        Xs_pos = ibackend.apply(keep_positives, Xs)
        Xs_neg = ibackend.apply(keep_negatives, Xs)
        # xpos*wpos + xneg*wneg
        activator_relevances = fn_tmp(
            self._layer_wo_act_positive, self._layer_wo_act_negative, Xs_pos, Xs_neg
        )

        if self._beta:  # only compute beta-weighted contributions of beta is not zero
            # xpos*wneg + xneg*wpos
            inhibitor_relevances = fn_tmp(
                self._layer_wo_act_negative, self._layer_wo_act_positive, Xs_pos, Xs_neg
            )
            return [
                klayers.Subtract()([times_alpha(a), times_beta(b)])
                for a, b in zip(activator_relevances, inhibitor_relevances)
            ]
        return activator_relevances


class AlphaBetaIgnoreBiasRule(AlphaBetaRule):
    """Same as AlphaBetaRule but ignores biases."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, bias=False, **kwargs)


class Alpha2Beta1Rule(AlphaBetaRule):
    """AlphaBetaRule with alpha=2, beta=1"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, alpha=2, beta=1, bias=True, **kwargs)


class Alpha2Beta1IgnoreBiasRule(AlphaBetaRule):
    """AlphaBetaRule with alpha=2, beta=1 and ignores biases"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, alpha=2, beta=1, bias=False, **kwargs)


class Alpha1Beta0Rule(AlphaBetaRule):
    """AlphaBetaRule with alpha=1, beta=0"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, alpha=1, beta=0, bias=True, **kwargs)


class Alpha1Beta0IgnoreBiasRule(AlphaBetaRule):
    """AlphaBetaRule with alpha=1, beta=0 and ignores biases"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, alpha=1, beta=0, bias=False, **kwargs)


class AlphaBetaXRule(igraph.ReverseMappingBase):
    """
    AlphaBeta advanced as proposed by Alexander Binder.
    """

    def __init__(
        self,
        layer: Layer,
        _state,
        alpha: tuple[float, float] = (0.5, 0.5),
        beta: tuple[float, float] = (0.5, 0.5),
        bias: bool = True,
        copy_weights: bool = False,
    ) -> None:
        self._alpha = alpha
        self._beta = beta

        # prepare positive and negative weights for computing positive
        # and negative preactivations z in apply_accordingly.
        if copy_weights:
            weights = layer.get_weights()
            if not bias and getattr(layer, "use_bias", False):
                weights = weights[:-1]
            positive_weights = [x * (x > 0) for x in weights]
            negative_weights = [x * (x < 0) for x in weights]
        else:
            weights = layer.weights
            if not bias and getattr(layer, "use_bias", False):
                weights = weights[:-1]
            positive_weights = [x * ibackend.cast_to_floatx(x > 0) for x in weights]
            negative_weights = [x * ibackend.cast_to_floatx(x < 0) for x in weights]

        self._layer_wo_act_positive = igraph.copy_layer_wo_activation(
            layer,
            keep_bias=bias,
            weights=positive_weights,
            name_template="reversed_kernel_positive_%s",
        )
        self._layer_wo_act_negative = igraph.copy_layer_wo_activation(
            layer,
            keep_bias=bias,
            weights=negative_weights,
            name_template="reversed_kernel_negative_%s",
        )

    def apply(
        self,
        Xs: list[Tensor],
        _Ys: list[Tensor],
        Rs: list[Tensor],
        _reverse_state: dict,
    ):
        # this method is correct, but wasteful
        times_alpha0 = klayers.Lambda(lambda x: x * self._alpha[0])
        # times_alpha1 = klayers.Lambda(lambda x: x * self._alpha[1]) # unused
        times_beta0 = klayers.Lambda(lambda x: x * self._beta[0])
        times_beta1 = klayers.Lambda(lambda x: x * self._beta[1])
        keep_positives = klayers.Lambda(
            lambda x: x * kbackend.cast(kbackend.greater(x, 0), kbackend.floatx())
        )
        keep_negatives = klayers.Lambda(
            lambda x: x * kbackend.cast(kbackend.less(x, 0), kbackend.floatx())
        )

        def fn_tmp(layer: Layer, Xs: OptionalList[Tensor]):
            Zs = ibackend.apply(layer, Xs)
            # Divide incoming relevance by the activations.
            tmp = [ibackend.safe_divide(a, b) for a, b in zip(Rs, Zs)]
            # Propagate the relevance to the input neurons
            # using the gradient
            grads = ibackend.gradients(Xs, Zs, tmp)
            # Re-weight relevance with the input values.
            return [klayers.Multiply()([a, b]) for a, b in zip(Xs, grads)]

        # Distinguish postive and negative inputs.
        Xs_pos = ibackend.apply(keep_positives, Xs)
        Xs_neg = ibackend.apply(keep_negatives, Xs)

        # xpos*wpos
        r_pp = fn_tmp(self._layer_wo_act_positive, Xs_pos)
        # xneg*wneg
        r_nn = fn_tmp(self._layer_wo_act_negative, Xs_neg)
        # a0 * r_pp + a1 * r_nn
        r_pos = [
            klayers.Add()([times_alpha0(pp), times_beta1(nn)])
            for pp, nn in zip(r_pp, r_nn)
        ]

        # xpos*wneg
        r_pn = fn_tmp(self._layer_wo_act_negative, Xs_pos)
        # xneg*wpos
        r_np = fn_tmp(self._layer_wo_act_positive, Xs_neg)
        # b0 * r_pn + b1 * r_np
        r_neg = [
            klayers.Add()([times_beta0(pn), times_beta1(np)])
            for pn, np in zip(r_pn, r_np)
        ]

        return [klayers.Subtract()([a, b]) for a, b in zip(r_pos, r_neg)]


class AlphaBetaX1000Rule(AlphaBetaXRule):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, alpha=(1, 0), beta=(0, 0), bias=True, **kwargs)


class AlphaBetaX1010Rule(AlphaBetaXRule):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, alpha=(1, 0), beta=(0, -1), bias=True, **kwargs)


class AlphaBetaX1001Rule(AlphaBetaXRule):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, alpha=(1, 1), beta=(0, 0), bias=True, **kwargs)


class AlphaBetaX2m100Rule(AlphaBetaXRule):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, alpha=(2, 0), beta=(1, 0), bias=True, **kwargs)


class BoundedRule(igraph.ReverseMappingBase):
    """Z_B rule from the Deep Taylor Decomposition"""

    # TODO: this only works for relu networks, needs to be extended.
    # TODO: check
    def __init__(
        self, layer: Layer, _state, low=-1, high=1, copy_weights: bool = False
    ) -> None:
        self._low = low
        self._high = high

        # This rule works with three variants of the layer, all without biases.
        # One is the original form and two with only the positive or
        # negative weights.
        if copy_weights:
            weights = layer.get_weights()
            if getattr(layer, "use_bias", False):
                weights = weights[:-1]
            positive_weights = [x * (x > 0) for x in weights]
            negative_weights = [x * (x < 0) for x in weights]
        else:
            weights = layer.weights
            if getattr(layer, "use_bias", False):
                weights = weights[:-1]
            positive_weights = [x * ibackend.cast_to_floatx(x > 0) for x in weights]
            negative_weights = [x * ibackend.cast_to_floatx(x < 0) for x in weights]

        self._layer_wo_act = igraph.copy_layer_wo_activation(
            layer, keep_bias=False, name_template="reversed_kernel_%s"
        )
        self._layer_wo_act_positive = igraph.copy_layer_wo_activation(
            layer,
            keep_bias=False,
            weights=positive_weights,
            name_template="reversed_kernel_positive_%s",
        )
        self._layer_wo_act_negative = igraph.copy_layer_wo_activation(
            layer,
            keep_bias=False,
            weights=negative_weights,
            name_template="reversed_kernel_negative_%s",
        )

    # TODO: clean up this implementation and add more documentation
    def apply(self, Xs, _Ys, Rs, reverse_state: dict):
        to_low = klayers.Lambda(lambda x: x * 0 + self._low)
        to_high = klayers.Lambda(lambda x: x * 0 + self._high)

        low = [to_low(x) for x in Xs]
        high = [to_high(x) for x in Xs]

        # Get values for the division.
        A = ibackend.apply(self._layer_wo_act, Xs)
        B = ibackend.apply(self._layer_wo_act_positive, low)
        C = ibackend.apply(self._layer_wo_act_negative, high)
        Zs = [
            klayers.Subtract()([a, klayers.Add()([b, c])]) for a, b, c in zip(A, B, C)
        ]

        # Divide relevances with the value.
        tmp = [ibackend.safe_divide(a, b) for a, b in zip(Rs, Zs)]
        # Distribute along the gradient.
        grads_a = ibackend.gradients(Xs, A, tmp)
        grads_b = ibackend.gradients(low, B, tmp)
        grads_c = ibackend.gradients(high, C, tmp)

        tmp_a = [klayers.Multiply()([a, b]) for a, b in zip(Xs, grads_a)]
        tmp_b = [klayers.Multiply()([a, b]) for a, b in zip(low, grads_b)]
        tmp_c = [klayers.Multiply()([a, b]) for a, b in zip(high, grads_c)]

        ret = [
            klayers.Subtract()([a, klayers.Add()([b, c])])
            for a, b, c in zip(tmp_a, tmp_b, tmp_c)
        ]

        return ret


class ZPlusRule(Alpha1Beta0IgnoreBiasRule):
    """
    The ZPlus rule is a special case of the AlphaBetaRule
    for alpha=1, beta=0, which assumes inputs x >= 0
    and ignores the bias.
    CAUTION! Results differ from Alpha=1, Beta=0
    if inputs are not strictly >= 0
    """

    # TODO: assert that layer inputs are always >= 0


class ZPlusFastRule(igraph.ReverseMappingBase):
    """
    The ZPlus rule is a special case of the AlphaBetaRule
    for alpha=1, beta=0 and assumes inputs x >= 0.
    """

    def __init__(self, layer: Layer, _state, copy_weights=False):
        # The z-plus rule only works with positive weights and
        # no biases.
        # TODO: assert that layer inputs are always >= 0
        if copy_weights:
            weights = layer.get_weights()
            if getattr(layer, "use_bias", False):
                weights = weights[:-1]
            weights = [x * (x > 0) for x in weights]
        else:
            weights = layer.weights
            if getattr(layer, "use_bias", False):
                weights = weights[:-1]
            weights = [x * ibackend.cast_to_floatx(x > 0) for x in weights]

        self._layer_wo_act_b_positive = igraph.copy_layer_wo_activation(
            layer,
            keep_bias=False,
            weights=weights,
            name_template="reversed_kernel_positive_%s",
        )

    def apply(self, Xs, _Ys, Rs, reverse_state: dict):
        # TODO: assert all inputs are positive, instead of only keeping the positives.
        # keep_positives = klayers.Lambda(
        #     lambda x: x * kbackend.cast(kbackend.greater(x, 0), kbackend.floatx())
        # )
        # Xs = ibackend.apply(keep_positives, Xs)

        # Get activations.
        Zs = ibackend.apply(self._layer_wo_act_b_positive, Xs)
        # Divide incoming relevance by the activations.
        tmp = [ibackend.safe_divide(a, b) for a, b in zip(Rs, Zs)]
        # Propagate the relevance to input neurons
        # using the gradient.
        grads = ibackend.gradients(Xs, Zs, tmp)
        # Re-weight relevance with the input values.
        return [klayers.Multiply()([a, b]) for a, b in zip(Xs, grads)]



from __future__ import annotations

import time
import warnings

import numpy as np
import tensorflow.keras.backend as kbackend
import tensorflow.keras.utils as kutils

import innvestigate.utils.sequence as isequence


class Perturbation:
    """Perturbation of pixels based on analysis result.

    :param perturbation_function:
        Defines the function with which the samples are perturbated.
        Can be a function or a string that defines a predefined perturbation function.
    :type perturbation_function:
        function or callable or str
    :param num_perturbed_regions:
        Number of regions to be perturbed.
    :type num_perturbed_regions:
        int
    :param reduce_function:
        Function to reduce the analysis result to one channel,
        e.g. mean or max function.
    :type reduce_function:
        function or callable
    :param aggregation_function:
        Function to aggregate the analysis over subregions.
    :type aggregation_function:
        function or callable
    :param pad_mode:
        How to pad if the image cannot be subdivided into an integer number of regions.
        As in numpy.pad.
    :type pad_mode:
        str or function or callable
    :param in_place:
        If true, the perturbations are performed in place,
        i.e. the input samples are modified.
    :type in_place:
        bool
    :param value_range:
        Minimal and maximal value after perturbation as a tuple:
        (min_val, max_val). The input is clipped to this range
    :type value_range: tuple"""

    def __init__(
        self,
        perturbation_function,
        num_perturbed_regions=0,
        region_shape=(9, 9),
        reduce_function=np.mean,
        aggregation_function=np.mean,
        pad_mode="reflect",
        in_place=False,
        value_range=None,
    ):
        if isinstance(perturbation_function, str):
            if perturbation_function == "zeros":
                # This is equivalent to setting the perturbated values
                # to the channel mean if the data are standardized.
                self.perturbation_function = np.zeros_like
            elif perturbation_function == "gaussian":
                # If scale = 1/3, most of the values will be between -1 and 1
                self.perturbation_function = lambda x: np.random.normal(
                    loc=0.0, scale=0.3, size=x.shape
                )
            elif perturbation_function == "mean":
                self.perturbation_function = np.mean
            elif perturbation_function == "invert":
                self.perturbation_function = lambda x: -x
            else:
                raise ValueError(
                    f"Perturbation function type '{perturbation_function}' not known."
                )
        elif callable(perturbation_function):
            self.perturbation_function = perturbation_function
        else:
            raise TypeError(
                f"Cannot handle perturbation function of type {perturbation_function}."
            )

        self.num_perturbed_regions = num_perturbed_regions
        self.region_shape = region_shape
        self.reduce_function = reduce_function
        self.aggregation_function = aggregation_function

        self.pad_mode = pad_mode  # numpy.pad

        self.in_place = in_place
        self.value_range = value_range

    @staticmethod
    def _compute_perturbation_mask(ranks, num_perturbated_regions):
        perturbation_mask_regions = ranks <= num_perturbated_regions - 1
        return perturbation_mask_regions

    @staticmethod
    def _compute_region_ordering(aggregated_regions):
        # 0 means highest scoring region
        new_shape = tuple(aggregated_regions.shape[:2]) + (-1,)
        order = np.argsort(-aggregated_regions.reshape(new_shape), axis=-1)
        ranks = order.argsort().reshape(aggregated_regions.shape)
        return ranks

    def _expand_regions_to_pixels(self, regions):
        # Resize to pixels (repeat values).
        # (n, c, h_aggregated_region, w_aggregated_region) ->
        # (n, c, h_aggregated_region, h_region, w_aggregated_region, w_region)
        regions_reshaped = np.expand_dims(np.expand_dims(regions, axis=3), axis=5)
        region_pixels = np.repeat(regions_reshaped, self.region_shape[0], axis=3)
        region_pixels = np.repeat(region_pixels, self.region_shape[1], axis=5)
        assert region_pixels.shape[0] == regions.shape[0] and region_pixels.shape[
            2:
        ] == (
            regions.shape[2],
            self.region_shape[0],
            regions.shape[3],
            self.region_shape[1],
        ), region_pixels.shape

        return region_pixels

    def _reshape_region_pixels(self, region_pixels, target_shape):
        # Reshape to output shape
        pixels = region_pixels.reshape(target_shape)
        assert (
            region_pixels.shape[0] == pixels.shape[0]
            and region_pixels.shape[1] == pixels.shape[1]
            and region_pixels.shape[2] * region_pixels.shape[3] == pixels.shape[2]
            and region_pixels.shape[4] * region_pixels.shape[5] == pixels.shape[3]
        )
        return pixels

    def _pad(self, analysis):
        pad_shape = self.region_shape - np.array(analysis.shape[2:]) % self.region_shape
        assert np.all(pad_shape < self.region_shape)

        # Pad half the window before and half after (on h and w axes)
        pad_shape_before = (pad_shape / 2).astype(int)
        pad_shape_after = pad_shape - pad_shape_before
        pad_shape = (
            (0, 0),
            (0, 0),
            (pad_shape_before[0], pad_shape_after[0]),
            (pad_shape_before[1], pad_shape_after[1]),
        )
        analysis = np.pad(analysis, pad_shape, self.pad_mode)
        assert np.all(
            np.array(analysis.shape[2:]) % self.region_shape == 0
        ), analysis.shape[2:]
        return analysis, pad_shape_before

    def _reshape_to_regions(self, analysis):
        aggregated_shape = tuple(
            (np.array(analysis.shape[2:]) / self.region_shape).astype(int)
        )
        regions = analysis.reshape(
            (
                analysis.shape[0],
                analysis.shape[1],
                aggregated_shape[0],
                self.region_shape[0],
                aggregated_shape[1],
                self.region_shape[1],
            )
        )
        return regions

    def _aggregate_regions(self, analysis):
        # Make sure color channel has been reduced:
        assert analysis.shape[1] == 1, analysis.shape

        regions = self._reshape_to_regions(analysis)
        aggregated_regions = self.aggregation_function(regions, axis=(3, 5))
        return aggregated_regions

    def _perturbate_regions(self, x, perturbation_mask_regions):
        # Perturbate every region in tensor.
        # A single region (at region_x, region_y in sample)
        # should be in mask[sample, channel, region_x, :, region_y, :]
        n_channels = x.shape[1]
        xp = self._reshape_to_regions(x)  # perturbed output
        for sample_idx, channel_idx, region_row, region_col in np.ndindex(
            perturbation_mask_regions.shape
        ):
            region_mask = perturbation_mask_regions[
                sample_idx, channel_idx, region_row, region_col
            ]
            if region_mask:
                for c in range(n_channels):
                    region = xp[sample_idx, c, region_row, :, region_col, :]
                    xp[
                        sample_idx, c, region_row, :, region_col, :
                    ] = self.perturbation_function(region)

        if self.value_range is not None:
            np.clip(xp, self.value_range[0], self.value_range[1], xp)
        xp = self._reshape_region_pixels(xp, x.shape)
        return xp

    def perturbate_on_batch(self, x, analysis):
        """
        :param x: Batch of images.
        :type x: numpy.ndarray
        :param analysis: Analysis of this batch.
        :type analysis: numpy.ndarray
        :return: Batch of perturbated images
        :rtype: numpy.ndarray
        """
        # Internally use channels_first (BCHW) only:
        if kbackend.image_data_format() == "channels_last":
            x = np.moveaxis(x, 3, 1)
            analysis = np.moveaxis(analysis, 3, 1)
        if not self.in_place:
            x = np.copy(x)
        assert analysis.shape == x.shape, analysis.shape
        original_shape = x.shape
        # reduce the analysis along channel axis -> n x 1 x h x w
        analysis = self.reduce_function(analysis, axis=1, keepdims=True)
        assert analysis.shape == (x.shape[0], 1, x.shape[2], x.shape[3]), analysis.shape

        padding = not np.all(np.array(analysis.shape[2:]) % self.region_shape == 0)
        if padding:
            analysis, _pad_shape_before_analysis = self._pad(analysis)
            x, pad_shape_before_x = self._pad(x)
        aggregated_regions = self._aggregate_regions(analysis)

        # Compute perturbation mask
        # (mask with ones where the input should be perturbated, zeros otherwise)
        ranks = self._compute_region_ordering(aggregated_regions)
        perturbation_mask_regions = self._compute_perturbation_mask(
            ranks, self.num_perturbed_regions
        )
        # Perturbate each region
        x_perturbated = self._perturbate_regions(x, perturbation_mask_regions)

        # Crop the original image region to remove the padding
        if padding:
            x_perturbated = x_perturbated[
                :,
                :,
                pad_shape_before_x[0] : pad_shape_before_x[0] + original_shape[2],
                pad_shape_before_x[1] : pad_shape_before_x[1] + original_shape[3],
            ]

        if kbackend.image_data_format() == "channels_last":
            x_perturbated = np.moveaxis(x_perturbated, 1, 3)
            x = np.moveaxis(x, 1, 3)
            analysis = np.moveaxis(analysis, 1, 3)
        return x_perturbated


class PerturbationAnalysis:
    """
    Performs the perturbation analysis.

    :param analyzer: Analyzer.
    :type analyzer: innvestigate.analyzer.base.AnalyzerBase
    :param model: Trained Keras model.
    :param generator: Data generator.
    :type generator: innvestigate.sequence.BatchSequence
    :param perturbation: Instance of Perturbation class that performs the perturbation.
    :type perturbation: innvestigate.tools.Perturbation
    :param steps: Number of perturbation steps.
    :type steps: int
    :param regions_per_step: Number of regions that are perturbed per step.
    :type regions_per_step: float
    :param recompute_analysis:
        If true, the analysis is recomputed after each perturbation step.
    :type recompute_analysis: bool
    :param verbose: If true, print some useful information, e.g. timing, progress etc.
    """

    def __init__(
        self,
        analyzer,
        model,
        generator,
        perturbation,
        steps=1,
        regions_per_step=1,
        recompute_analysis=False,
        verbose=False,
    ):
        self.analyzer = analyzer
        self.model = model
        self.generator = generator
        self.perturbation = perturbation
        # if not isinstance(perturbation, Perturbation):
        #     raise TypeError(type(perturbation))
        self.steps = steps
        self.regions_per_step = regions_per_step
        self.recompute_analysis = recompute_analysis

        if not self.recompute_analysis:
            # Compute the analysis once in the beginning
            analysis = []
            X = []
            Y = []
            for XX, YY in self.generator:
                X.extend(list(XX))
                Y.extend(list(YY))
                analysis.extend(list(self.analyzer.analyze(XX)))
            X = np.array(X)
            Y = np.array(Y)
            analysis = np.array(analysis)
            self.analysis_generator = isequence.BatchSequence(
                [X, Y, analysis], batch_size=256
            )
        self.verbose = verbose

    def compute_on_batch(self, x, analysis=None, return_analysis=False):
        """
        Computes the analysis and perturbes the input batch accordingly.

        :param x: Samples.
        :param analysis: Analysis of x. If None, it is recomputed.
        :type x: numpy.ndarray
        """
        if analysis is None:
            analysis = self.analyzer.analyze(x)

        x_perturbated = self.perturbation.perturbate_on_batch(x, analysis)
        if return_analysis:
            return x_perturbated, analysis
        return x_perturbated

    def evaluate_on_batch(self, x, y, analysis=None, sample_weight=None):
        """
        Perturbs the input batch and scores the model on the perturbed batch.

        :param x: Samples.
        :type x: numpy.ndarray
        :param y: Labels.
        :type y: numpy.ndarray
        :param analysis: Analysis of x.
        :type analysis: numpy.ndarray
        :param sample_weight: Sample weights.
        :type sample_weight: None
        :return: List of test scores.
        :rtype: list
        """
        if sample_weight is not None:
            raise NotImplementedError("Sample weighting is not supported yet.")  # TODO
        x_perturbated = self.compute_on_batch(x, analysis)
        score = self.model.test_on_batch(x_perturbated, y, sample_weight=sample_weight)
        return score

    def evaluate_generator(
        self,
        generator,
        steps=None,
        max_queue_size=10,
        workers=1,
        use_multiprocessing=False,
    ):
        """Evaluates the model on a data generator.

        The generator should return the same kind of data
        as accepted by `test_on_batch`.
        For documentation, refer to keras.engine.training.evaluate_generator
        (https://keras.io/models/model/)
        """

        steps_done = 0
        all_outs = []
        batch_sizes = []
        is_sequence = isinstance(generator, kutils.Sequence)
        if not is_sequence and use_multiprocessing and workers > 1:
            warnings.warn(
                UserWarning(
                    "Using a generator with `use_multiprocessing=True`"
                    " and multiple workers may duplicate your data."
                    " Please consider using the`keras.utils.Sequence"
                    " class."
                )
            )
        if steps is None:
            if is_sequence:
                steps = len(generator)
            else:
                raise ValueError(
                    "`steps=None` is only valid for a generator"
                    " based on the `keras.utils.Sequence` class."
                    " Please specify `steps` or use the"
                    " `keras.utils.Sequence` class."
                )
        enqueuer = None

        try:
            if workers > 0:
                if is_sequence:
                    enqueuer = kutils.OrderedEnqueuer(
                        generator, use_multiprocessing=use_multiprocessing
                    )
                else:
                    enqueuer = kutils.GeneratorEnqueuer(
                        generator,
                        use_multiprocessing=use_multiprocessing,
                    )
                enqueuer.start(workers=workers, max_queue_size=max_queue_size)
                output_generator = enqueuer.get()
            else:
                output_generator = generator

            while steps_done < steps:
                generator_output = next(output_generator)
                if not hasattr(generator_output, "__len__"):
                    raise ValueError(
                        "Output of generator should be a tuple "
                        "(x, y, sample_weight) "
                        "or (x, y). Found: " + str(generator_output)
                    )
                if len(generator_output) == 2:
                    x, y = generator_output
                    analysis = None
                elif len(generator_output) == 3:
                    x, y, analysis = generator_output
                else:
                    raise ValueError(
                        "Output of generator should be a tuple "
                        "(x, y, analysis) "
                        "or (x, y). Found: " + str(generator_output)
                    )
                outs = self.evaluate_on_batch(
                    x, y, analysis=analysis, sample_weight=None
                )

                if isinstance(x, list):
                    batch_size = x[0].shape[0]
                elif isinstance(x, dict):
                    batch_size = list(x.values())[0].shape[0]
                else:
                    batch_size = x.shape[0]
                if batch_size == 0:
                    raise ValueError(
                        "Received an empty batch. "
                        "Batches should at least contain one item."
                    )
                all_outs.append(outs)

                steps_done += 1
                batch_sizes.append(batch_size)

        finally:
            if enqueuer is not None:
                enqueuer.stop()

        if not isinstance(outs, list):
            return np.average(np.asarray(all_outs), weights=batch_sizes)
        averages = []
        for i in range(len(outs)):
            averages.append(
                np.average([out[i] for out in all_outs], weights=batch_sizes)
            )
        return averages

    def compute_perturbation_analysis(self):
        scores = []
        # Evaluate first on original data
        scores.append(self.model.evaluate(self.generator))
        self.perturbation.num_perturbed_regions = 1
        time_start = time.time()
        for step in range(self.steps):
            tic = time.time()
            if self.verbose:
                print(
                    f"Step {step + 1} of {self.steps}: "
                    f"{self.perturbation.num_perturbed_regions} regions perturbed.",
                    end=" ",
                )
            scores.append(self.evaluate_generator(self.analysis_generator))
            self.perturbation.num_perturbed_regions += self.regions_per_step
            toc = time.time()
            if self.verbose:
                print(f"Time elapsed: {toc - tic:.3f} seconds.")
        time_end = time.time()

        if self.verbose:
            print(
                # Use step + 1 instead of self.steps
                # because the analysis can stop prematurely
                f"Time elapsed for {step + 1} steps: "
                f"{time_end - time_start:.3f} seconds."
            )

        self.perturbation.num_perturbed_regions = 1  # Reset to original value
        assert len(scores) == self.steps + 1
        return scores



from __future__ import annotations

import numpy as np
import tensorflow.keras.backend as kbackend
import tensorflow.keras.layers as klayers
import tensorflow.keras.models as kmodels

import innvestigate.backend as ibackend
import innvestigate.layers as ilayers
from innvestigate.analyzer.base import AnalyzerBase
from innvestigate.analyzer.network_base import AnalyzerNetworkBase
from innvestigate.backend.types import OptionalList, Tensor

__all__ = [
    "WrapperBase",
    "AugmentReduceBase",
    "GaussianSmoother",
    "PathIntegrator",
]


class WrapperBase(AnalyzerBase):
    """Interface for wrappers around analyzers

    This class is the basic interface for wrappers around analyzers.

    :param subanalyzer: The analyzer to be wrapped.
    """

    def __init__(self, subanalyzer: AnalyzerBase, *args, **kwargs):
        if not isinstance(subanalyzer, AnalyzerNetworkBase):
            raise NotImplementedError("Keras-based subanalyzer required.")

        # To simplify serialization, additionaly passed models are popped
        # and the subanalyzer model is passed to `AnalyzerBase`.
        kwargs.pop("model", None)
        super().__init__(subanalyzer._model, *args, **kwargs)

        self._subanalyzer_name = subanalyzer.__class__.__name__
        self._subanalyzer = subanalyzer

    def analyze(self, *args, **kwargs):
        return self._subanalyzer.analyze(*args, **kwargs)

    def _get_state(self) -> dict:
        sa_class_name, sa_state = self._subanalyzer.save()

        state = super()._get_state()
        state.update({"subanalyzer_class_name": sa_class_name})
        state.update({"subanalyzer_state": sa_state})
        return state

    @classmethod
    def _state_to_kwargs(cls, state: dict):
        sa_class_name = state.pop("subanalyzer_class_name")
        sa_state = state.pop("subanalyzer_state")
        # call super after popping class-specific states:
        kwargs = super()._state_to_kwargs(state)

        subanalyzer = AnalyzerBase.load(sa_class_name, sa_state)
        kwargs.update({"subanalyzer": subanalyzer})
        return kwargs


###############################################################################


class AugmentReduceBase(WrapperBase):
    """Interface for wrappers that augment the input and reduce the analysis.

    This class is an interface for wrappers that:
    * augment the input to the analyzer by creating new samples.
    * reduce the returned analysis to match the initial input shapes.

    :param subanalyzer: The analyzer to be wrapped.
    :param augment_by_n: Number of samples to create.
    """

    def __init__(
        self,
        subanalyzer: AnalyzerNetworkBase,
        *args,
        augment_by_n: int = 2,
        neuron_selection_mode="max_activation",
        **kwargs,
    ):
        if neuron_selection_mode == "max_activation":
            # TODO: find a more transparent way.
            #
            # Since AugmentReduceBase analyzers augment the input,
            # it is possible that the neuron w/ max activation changes.
            # As a workaround, the index of the maximally activated neuron
            # w.r.t. the "unperturbed" input is computed and used in combination
            # with neuron_selection_mode = "index" in the subanalyzer.
            #
            # NOTE:
            # The analyzer will still have neuron_selection_mode = "max_activation"!
            subanalyzer._neuron_selection_mode = "index"

        super().__init__(
            subanalyzer, *args, neuron_selection_mode=neuron_selection_mode, **kwargs
        )

        self._augment_by_n: int = augment_by_n  # number of samples to create

    def create_analyzer_model(self):
        self._subanalyzer.create_analyzer_model()

        if self._subanalyzer._n_debug_output > 0:
            raise NotImplementedError("No debug output at subanalyzer is supported.")

        model = self._subanalyzer._analyzer_model
        if None in model.input_shape[1:]:
            raise ValueError(
                "The input shape for the model needs "
                "to be fully specified (except the batch axis). "
                f"Model input shape is: {model.input_shape}"
            )

        inputs = model.inputs[: self._subanalyzer._n_data_input]
        extra_inputs = model.inputs[self._subanalyzer._n_data_input :]

        # outputs = model.outputs[: self._subanalyzer._n_data_output]
        extra_outputs = model.outputs[self._subanalyzer._n_data_output :]

        if len(extra_outputs) > 0:
            raise Exception("No extra output is allowed with this wrapper.")

        new_inputs = self._augment(ibackend.to_list(inputs))

        augmented_outputs = ibackend.to_list(model(new_inputs + extra_inputs))
        new_outputs = self._reduce(augmented_outputs)
        new_constant_inputs = self._keras_get_constant_inputs()

        inputs = inputs + extra_inputs + new_constant_inputs
        outputs = new_outputs + extra_outputs
        new_model = kmodels.Model(inputs=inputs, outputs=outputs)
        self._subanalyzer._analyzer_model = new_model

    def analyze(
        self, X: OptionalList[np.ndarray], *args, **kwargs
    ) -> OptionalList[np.ndarray]:
        if self._subanalyzer._analyzer_model is None:
            self.create_analyzer_model()

        ns_mode = self._neuron_selection_mode
        if ns_mode == "all":
            return self._subanalyzer.analyze(X, *args, **kwargs)

        # As described in the AugmentReduceBase init,
        # both ns_mode "max_activation" and "index" make use
        # of a subanalyzer using neuron_selection_mode="index".
        if ns_mode == "max_activation":
            # obtain max neuron activations over batch
            pred = self._subanalyzer._model.predict(X)
            indices = np.argmax(pred, axis=1)
        elif ns_mode == "index":
            # TODO: make neuron_selection arg or kwarg, not both
            if args:
                indices = list(args).pop(0)
            else:
                indices = kwargs.pop("neuron_selection")

        if not self._subanalyzer._neuron_selection_mode == "index":
            raise AssertionError(
                'Subanalyzer neuron_selection_mode has to be "index" '
                'when using analyzer with neuron_selection_mode != "all".'
            )
        # broadcast to match augmented samples.
        indices = np.repeat(indices, self._augment_by_n)
        kwargs["neuron_selection"] = indices
        return self._subanalyzer.analyze(X, *args, **kwargs)

    def _keras_get_constant_inputs(self) -> list[Tensor] | None:
        return []

    def _augment(self, Xs: OptionalList[Tensor]) -> list[Tensor]:
        """Augment inputs before analyzing them with subanalyzer."""
        repeat = ilayers.Repeat(self._augment_by_n)
        reshape = ilayers.AugmentationToBatchAxis(self._augment_by_n)
        return [reshape(repeat(X)) for X in ibackend.to_list(Xs)]

    def _reduce(self, Xs: OptionalList[Tensor]) -> list[Tensor]:
        """Reduce input Xs by reshaping and taking the mean along
        the axis of augmentation."""
        reshape = ilayers.AugmentationFromBatchAxis(self._augment_by_n)
        reduce = ilayers.ReduceMean()
        means = [reduce(reshape(X)) for X in ibackend.to_list(Xs)]
        return means

    def _get_state(self):
        state = super()._get_state()
        state.update({"augment_by_n": self._augment_by_n})
        return state

    @classmethod
    def _state_to_kwargs(cls, state):
        augment_by_n = state.pop("augment_by_n")
        # call super after popping class-specific states:
        kwargs = super()._state_to_kwargs(state)

        kwargs.update({"augment_by_n": augment_by_n})
        return kwargs


###############################################################################


class GaussianSmoother(AugmentReduceBase):
    """Wrapper that adds noise to the input and averages over analyses

    This wrapper creates new samples by adding Gaussian noise
    to the input. The final analysis is an average of the returned analyses.

    :param subanalyzer: The analyzer to be wrapped.
    :param noise_scale: The stddev of the applied noise.
    :param augment_by_n: Number of samples to create.
    """

    def __init__(self, subanalyzer, *args, noise_scale: float = 1, **kwargs):
        super().__init__(subanalyzer, *args, **kwargs)
        self._noise_scale = noise_scale

    def _augment(self, Xs: OptionalList[Tensor]) -> list[Tensor]:
        repeat = ilayers.Repeat(self._augment_by_n)
        add_noise = ilayers.AddGaussianNoise()
        reshape = ilayers.AugmentationToBatchAxis(self._augment_by_n)

        ret = [reshape(add_noise(repeat(X))) for X in ibackend.to_list(Xs)]
        return ret

    def _get_state(self):
        state = super()._get_state()
        state.update({"noise_scale": self._noise_scale})
        return state

    @classmethod
    def _state_to_kwargs(cls, state):
        noise_scale = state.pop("noise_scale")
        # call super after popping class-specific states:
        kwargs = super()._state_to_kwargs(state)

        kwargs.update({"noise_scale": noise_scale})
        return kwargs


###############################################################################


class PathIntegrator(AugmentReduceBase):
    """Integrated the analysis along a path

    This analyzer:
    * creates a path from input to reference image.
    * creates `steps` number of intermediate inputs and
      creates an analysis for them.
    * sums the analyses and multiplies them with the input-reference_input.

    This wrapper is used to implement Integrated Gradients.
    We refer to the paper for further information.

    :param subanalyzer: The analyzer to be wrapped.
    :param steps: Number of steps for integration.
    :param reference_inputs: The reference input.
    """

    def __init__(
        self, subanalyzer, *args, steps: int = 16, reference_inputs=0, **kwargs
    ):
        super().__init__(subanalyzer, *args, augment_by_n=steps, **kwargs)

        self._reference_inputs = reference_inputs
        self._keras_constant_inputs: list[Tensor] | None = None

    def _keras_set_constant_inputs(self, inputs: list[Tensor]) -> None:
        tmp = [kbackend.variable(X) for X in inputs]
        self._keras_constant_inputs = [
            klayers.Input(tensor=X, shape=X.shape[1:]) for X in tmp
        ]

    def _keras_get_constant_inputs(self) -> list[Tensor] | None:
        return self._keras_constant_inputs

    def _compute_difference(self, Xs: list[Tensor]) -> list[Tensor]:
        if self._keras_constant_inputs is None:
            inputs = ibackend.broadcast_np_tensors_to_keras_tensors(
                self._reference_inputs, Xs
            )
            self._keras_set_constant_inputs(inputs)

        # Type not Optional anymore as as `_keras_set_constant_inputs` has been called.
        reference_inputs: list[Tensor]
        reference_inputs = self._keras_get_constant_inputs()  # type: ignore
        return [klayers.subtract([x, ri]) for x, ri in zip(Xs, reference_inputs)]

    def _augment(self, Xs):
        difference = self._compute_difference(Xs)
        self._keras_difference = difference
        # Make broadcastable.
        difference = [
            ilayers.Reshape((-1, 1) + kbackend.int_shape(x)[1:])(x) for x in difference
        ]

        # Compute path steps.
        multiply_with_linspace = ilayers.MultiplyWithLinspace(
            0, 1, n=self._augment_by_n, axis=1
        )
        path_steps = [multiply_with_linspace(d) for d in difference]

        reference_inputs = self._keras_get_constant_inputs()
        ret = [klayers.Add()([x, p]) for x, p in zip(reference_inputs, path_steps)]
        ret = [ilayers.Reshape((-1,) + kbackend.int_shape(x)[2:])(x) for x in ret]
        return ret

    def _reduce(self, Xs):
        tmp = super()._reduce(Xs)
        difference = self._keras_difference
        del self._keras_difference

        return [klayers.Multiply()([x, d]) for x, d in zip(tmp, difference)]

    def _get_state(self):
        state = super()._get_state()
        state.update({"reference_inputs": self._reference_inputs})
        return state

    @classmethod
    def _state_to_kwargs(cls, state):
        reference_inputs = state.pop("reference_inputs")
        # call super after popping class-specific states:
        kwargs = super()._state_to_kwargs(state)

        # We use steps instead.
        kwargs.update(
            {"reference_inputs": reference_inputs, "steps": kwargs["augment_by_n"]}
        )
        del kwargs["augment_by_n"]
        return kwargs



from __future__ import annotations

import tensorflow as tf
import tensorflow.keras.backend as kbackend
import tensorflow.keras.layers as klayers

import innvestigate.backend as ibackend
import innvestigate.backend.checks as ichecks
import innvestigate.backend.graph as igraph
from innvestigate.analyzer.network_base import AnalyzerNetworkBase
from innvestigate.analyzer.reverse_base import ReverseAnalyzerBase
from innvestigate.analyzer.wrapper import GaussianSmoother, PathIntegrator
from innvestigate.backend.types import List, OptionalList, Tensor

__all__ = [
    "BaselineGradient",
    "Gradient",
    "InputTimesGradient",
    "Deconvnet",
    "GuidedBackprop",
    "IntegratedGradients",
    "SmoothGrad",
]


class BaselineGradient(AnalyzerNetworkBase):
    """Gradient analyzer based on build-in gradient.

    Returns as analysis the function value with respect to the input.
    The gradient is computed via the build in function.
    Is mainly used for debugging purposes.

    :param model: A Keras model.
    """

    def __init__(self, model, postprocess=None, **kwargs):
        super().__init__(model, **kwargs)

        if postprocess not in [None, "abs", "square"]:
            raise ValueError(
                "Parameter 'postprocess' must be either " "None, 'abs', or 'square'."
            )
        self._postprocess = postprocess

        self._add_model_softmax_check()
        self._do_model_checks()

    def _create_analysis(self, model, stop_analysis_at_tensors=None):
        if stop_analysis_at_tensors is None:
            stop_analysis_at_tensors = []

        tensors_to_analyze = [
            x
            for x in ibackend.to_list(model.inputs)
            if x not in stop_analysis_at_tensors
        ]
        ret = ibackend.to_list(kbackend.gradients(model.outputs[0], tensors_to_analyze))

        if self._postprocess == "abs":
            ret = [kbackend.abs(r) for r in ret]
        elif self._postprocess == "square":
            ret = [kbackend.square(r) for r in ret]

        return ret

    def _get_state(self):
        state = super()._get_state()
        state.update({"postprocess": self._postprocess})
        return state

    @classmethod
    def _state_to_kwargs(cls, state):
        postprocess = state.pop("postprocess")
        # call super after popping class-specific states:
        kwargs = super()._state_to_kwargs(state)

        kwargs.update(
            {
                "postprocess": postprocess,
            }
        )
        return kwargs


class Gradient(ReverseAnalyzerBase):
    """Gradient analyzer.

    Returns as analysis the function value with respect to the input.
    The gradient is computed via the librarie's network reverting.

    :param model: A Keras model.
    """

    def __init__(self, model, postprocess: str | None = None, **kwargs):
        super().__init__(model, **kwargs)

        if postprocess not in [None, "abs", "square"]:
            raise ValueError(
                """Parameter 'postprocess' must be either None, "abs", or "square"."""
            )
        self._postprocess = postprocess

        # Add and run model checks
        self._add_model_softmax_check()
        self._do_model_checks()

    def _head_mapping(self, X: Tensor) -> Tensor:
        return tf.ones_like(X)

    def _postprocess_analysis(self, Xs: OptionalList[Tensor]) -> List[Tensor]:
        ret = super()._postprocess_analysis(Xs)

        if self._postprocess == "abs":
            ret = [kbackend.abs(r) for r in ret]
        elif self._postprocess == "square":
            ret = [kbackend.square(r) for r in ret]

        return ret

    def _get_state(self):
        state = super()._get_state()
        state.update({"postprocess": self._postprocess})
        return state

    @classmethod
    def _state_to_kwargs(cls, state):
        postprocess = state.pop("postprocess")
        # call super after popping class-specific states:
        kwargs = super()._state_to_kwargs(state)

        kwargs.update(
            {
                "postprocess": postprocess,
            }
        )
        return kwargs


###############################################################################


class InputTimesGradient(Gradient):
    """Input*Gradient analyzer.

    :param model: A Keras model.
    """

    def __init__(self, model, **kwargs):
        super().__init__(model, **kwargs)

    def _create_analysis(self, model, stop_analysis_at_tensors=None):
        if stop_analysis_at_tensors is None:
            stop_analysis_at_tensors = []

        tensors_to_analyze = [
            x
            for x in ibackend.to_list(model.inputs)
            if x not in stop_analysis_at_tensors
        ]
        gradients = super()._create_analysis(
            model, stop_analysis_at_tensors=stop_analysis_at_tensors
        )
        return [
            klayers.Multiply()([i, g]) for i, g in zip(tensors_to_analyze, gradients)
        ]


###############################################################################


class DeconvnetReverseReLULayer(igraph.ReverseMappingBase):
    def __init__(self, layer, _state):
        self._activation = klayers.Activation("relu")
        self._layer_wo_relu = igraph.copy_layer_wo_activation(
            layer,
            name_template="reversed_%s",
        )

    def apply(self, Xs, Ys, Rs, reverse_state: dict) -> List[Tensor]:
        # Apply relus conditioned on backpropagated values.
        Rs = ibackend.apply(self._activation, Rs)

        # Apply gradient of forward pass without relus.
        Ys_wo_relu = ibackend.apply(self._layer_wo_relu, Xs)
        return ibackend.gradients(Xs, Ys_wo_relu, Rs)


class Deconvnet(ReverseAnalyzerBase):
    """Deconvnet analyzer.

    Applies the "deconvnet" algorithm to analyze the model.

    :param model: A Keras model.
    """

    def __init__(self, model, **kwargs):
        super().__init__(model, **kwargs)

        # Add and run model checks
        self._add_model_softmax_check()
        self._add_model_check(
            lambda layer: not ichecks.only_relu_activation(layer),
            "Deconvnet is only specified for networks with ReLU activations.",
            check_type="exception",
        )
        self._do_model_checks()

    def _create_analysis(self, *args, **kwargs):
        self._add_conditional_reverse_mapping(
            lambda layer: ichecks.contains_activation(layer, "relu"),
            DeconvnetReverseReLULayer,
            name="deconvnet_reverse_relu_layer",
        )

        return super()._create_analysis(*args, **kwargs)


def guided_backprop_reverse_relu_layer(Xs, Ys, reversed_Ys, _reverse_state: dict):
    activation = klayers.Activation("relu")
    # Apply relus conditioned on backpropagated values.
    reversed_Ys = ibackend.apply(activation, reversed_Ys)

    # Apply gradient of forward pass.
    return ibackend.gradients(Xs, Ys, reversed_Ys)


class GuidedBackprop(ReverseAnalyzerBase):
    """Guided backprop analyzer.

    Applies the "guided backprop" algorithm to analyze the model.

    :param model: A Keras model.
    """

    def __init__(self, model, **kwargs):
        super().__init__(model, **kwargs)

        # Add and run model checks
        self._add_model_softmax_check()
        self._add_model_check(
            lambda layer: not ichecks.only_relu_activation(layer),
            "GuidedBackprop is only specified for networks with ReLU activations.",
            check_type="exception",
        )
        self._do_model_checks()

    def _create_analysis(self, *args, **kwargs):
        self._add_conditional_reverse_mapping(
            lambda layer: ichecks.contains_activation(layer, "relu"),
            guided_backprop_reverse_relu_layer,
            name="guided_backprop_reverse_relu_layer",
        )

        return super()._create_analysis(*args, **kwargs)


###############################################################################


class IntegratedGradients(PathIntegrator):
    """Integrated gradient analyzer.

    Applies the "integrated gradient" algorithm to analyze the model.

    :param model: A Keras model.
    :param steps: Number of steps to use average along integration path.
    """

    def __init__(
        self,
        model,
        steps=64,
        neuron_selection_mode="max_activation",
        postprocess=None,
        **kwargs,
    ):
        # If initialized through serialization:
        if "subanalyzer" in kwargs:
            subanalyzer = kwargs.pop("subanalyzer")
        # If initialized normally:
        else:
            subanalyzer = Gradient(
                model,
                neuron_selection_mode=neuron_selection_mode,
                postprocess=postprocess,
            )

        super().__init__(
            subanalyzer,
            steps=steps,
            neuron_selection_mode=neuron_selection_mode,
            **kwargs,
        )


###############################################################################


class SmoothGrad(GaussianSmoother):
    """Smooth grad analyzer.

    Applies the "smooth grad" algorithm to analyze the model.

    :param model: A Keras model.
    :param augment_by_n: Number of distortions to average for smoothing.
    """

    def __init__(
        self,
        model,
        augment_by_n=64,
        neuron_selection_mode="max_activation",
        postprocess=None,
        **kwargs,
    ):
        # If initialized through serialization:
        if "subanalyzer" in kwargs:
            subanalyzer = kwargs.pop("subanalyzer")
        # If initialized normally:
        else:
            subanalyzer = Gradient(
                model,
                neuron_selection_mode=neuron_selection_mode,
                postprocess=postprocess,
            )

        super().__init__(
            subanalyzer,
            augment_by_n=augment_by_n,
            neuron_selection_mode=neuron_selection_mode,
            **kwargs,
        )



from __future__ import annotations

from typing import Sequence

import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as kbackend
import tensorflow.keras.layers as klayers

import innvestigate.backend as ibackend
from innvestigate.backend.types import OptionalList, ShapeTuple, Tensor

__all__ = [
    "OnesLike",
    "AsFloatX",
    "FiniteCheck",
    "GreaterThanZero",
    "LessEqualThanZero",
    "Sum",
    "Identity",
    "Abs",
    "Square",
    "Clip",
    "Project",
    "SafeDivide",
    "Repeat",
    "ReduceMean",
    "Reshape",
    "AugmentationToBatchAxis",
    "AugmentationFromBatchAxis",
    "MultiplyWithLinspace",
    "AddGaussianNoise",
    "ExtractConv2DPatches",
    "RunningMeans",
    "Broadcast",
    "MaxNeuronSelection",
    "MaxNeuronIndex",
    "NeuronSelection",
]


###############################################################################


class OnesLike(klayers.Layer):
    """Create list of all-ones tensors of the same shapes as provided tensors."""

    def call(self, inputs: OptionalList[Tensor], *_args, **_kwargs) -> list[Tensor]:
        return [kbackend.ones_like(x) for x in ibackend.to_list(inputs)]


class AsFloatX(klayers.Layer):
    def call(self, inputs: OptionalList[Tensor], *_args, **_kwargs) -> list[Tensor]:
        return [ibackend.cast_to_floatx(x) for x in ibackend.to_list(inputs)]


class FiniteCheck(klayers.Layer):
    def call(self, inputs: OptionalList[Tensor], *_args, **_kwargs) -> list[Tensor]:
        return [
            kbackend.sum(ibackend.cast_to_floatx(ibackend.is_not_finite(x)))
            for x in ibackend.to_list(inputs)
        ]


###############################################################################


class _Reduce(klayers.Layer):
    def __init__(
        self,
        *args,
        axis: OptionalList[int] | None = -1,
        keepdims: bool = False,
        **kwargs,
    ) -> None:
        self.axis = axis
        self.keepdims = keepdims
        super().__init__(*args, **kwargs)

    def call(self, inputs: OptionalList[Tensor], *_args, **_kwargs) -> Tensor:
        return self._apply_reduce(inputs, axis=self.axis, keepdims=self.keepdims)

    def _apply_reduce(
        self,
        inputs: OptionalList[Tensor],
        axis: OptionalList[int] | None,
        keepdims: bool,
    ) -> Tensor:
        raise NotImplementedError()


class Sum(_Reduce):
    def _apply_reduce(
        self,
        inputs: OptionalList[Tensor],
        axis: OptionalList[int] | None,
        keepdims: bool,
    ) -> Tensor:
        return kbackend.sum(inputs, axis=axis, keepdims=keepdims)


###############################################################################


class _Map(klayers.Layer):
    def call(
        self, inputs: OptionalList[Tensor], *_args, **_kwargs
    ) -> OptionalList[Tensor]:
        if isinstance(inputs, list) and len(inputs) == 1:
            inputs = inputs[0]
        return self._apply_map(inputs)

    def _apply_map(self, X: Tensor):
        raise NotImplementedError()


class Identity(_Map):
    def _apply_map(self, X: Tensor) -> Tensor:
        return tf.identity(X)


class Abs(_Map):
    def _apply_map(self, X: Tensor) -> Tensor:
        return kbackend.abs(X)


class Square(_Map):
    def _apply_map(self, X: Tensor) -> Tensor:
        return kbackend.square(X)


class Clip(_Map):
    def __init__(
        self, min_value: float | int | Tensor, max_value: float | int | Tensor
    ) -> None:
        self._min_value = min_value
        self._max_value = max_value
        super().__init__()

    def _apply_map(self, X: Tensor) -> Tensor:
        return kbackend.clip(X, self._min_value, self._max_value)


class Project(_Map):
    def __init__(self, output_range=False, input_is_postive_only: bool = False) -> None:
        # TODO: add type of output_range
        self._output_range = output_range
        self._input_is_positive_only = input_is_postive_only
        super().__init__()

    def _apply_map(self, X: Tensor):
        dims: tuple[int] = kbackend.int_shape(X)
        n_dim: int = len(dims)
        axes = tuple(range(1, n_dim))

        if len(axes) == 1:
            # TODO(albermax): this is only the case when the dimension in this
            # axis is 1, fix this.
            # Cannot reduce
            return X

        absmax = kbackend.max(kbackend.abs(X), axis=axes, keepdims=True)
        X = ibackend.safe_divide(X, absmax, factor=1)

        if self._output_range not in (False, True):  # True = (-1, +1)
            output_range = self._output_range

            if not self._input_is_positive_only:
                X = (X + 1) / 2
            X = kbackend.clip(X, 0, 1)

            X = output_range[0] + (X * (output_range[1] - output_range[0]))
        else:
            X = kbackend.clip(X, -1, 1)

        return X


###############################################################################


class GreaterThanZero(klayers.Layer):
    def call(self, inputs: Tensor, *_args, **_kwargs) -> Tensor:
        return kbackend.greater(inputs, kbackend.constant(0))


class LessEqualThanZero(klayers.Layer):
    def call(self, inputs: Tensor, *_args, **_kwargs) -> Tensor:
        return kbackend.less_equal(inputs, kbackend.constant(0))


class Divide(klayers.Layer):
    def call(self, inputs: list[Tensor], *_args, **_kwargs) -> Tensor:
        if len(inputs) != 2:
            raise ValueError("A `Divide` layer should be called on exactly 2 inputs")
        a, b = inputs
        return a / b


class SafeDivide(klayers.Layer):
    def __init__(self, *args, factor: float = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        if factor is None:
            factor = kbackend.epsilon()
        self._factor = factor

    def call(self, inputs: list[Tensor], *_args, **_kwargs) -> Tensor:
        if len(inputs) != 2:
            raise ValueError(
                "A `SafeDivide` layer should be called on exactly 2 inputs"
            )
        a, b = inputs
        return ibackend.safe_divide(a, b, factor=self._factor)


###############################################################################
class Repeat(klayers.Layer):
    """Repeats the input n times. Similar to Keras' `RepeatVector`,
    except that it works on any Tensor.

    Input shape: 2D tensor of shape `(num_samples, features)`.
    Output shape: 3D tensor of shape `(num_samples, n, features)`.

    Args:
        n: Integer, repetition factor.
    """

    def __init__(self, n, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.n = n
        if not isinstance(n, int):
            raise TypeError(f"Expected an integer value for `n`, got {type(n)}.")

    def call(self, inputs, *_args, **_kwargs):
        dims = inputs.shape.rank  # number of axes in Tensor
        assert dims >= 2
        inputs = tf.expand_dims(inputs, 1)

        # Construct array [1, n, 1, ..., 1] for tf.tile
        multiples = [1] * dims
        multiples.insert(1, self.n)
        return tf.tile(inputs, tf.constant(multiples))


class ReduceMean(klayers.Layer):
    """Reduce input augmented along `axis=1` by taking the mean."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def call(self, inputs: Tensor, *_args, **_kwargs) -> Tensor:
        return tf.math.reduce_mean(inputs, axis=1, keepdims=False)


class Reshape(klayers.Layer):
    """Layer that reshapes tensor to the shape specified on init."""

    def __init__(self, shape: ShapeTuple, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._shape = shape

    def call(self, inputs: Tensor, *_args, **_kwargs) -> Tensor:
        return tf.reshape(inputs, self._shape)


class AugmentationToBatchAxis(klayers.Layer):
    """Move augmentation from axis=1 to batch axis."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def call(self, inputs: Tensor, *_args, **_kwargs) -> Tensor:
        input_shape = ibackend.shape(inputs)
        output_shape = [-1] + input_shape[2:]  # type: ignore
        return tf.reshape(inputs, output_shape)


class AugmentationFromBatchAxis(klayers.Layer):
    """Move augmentation from batch axis to axis=1.

    Args:
        n: Factor of augmentation.
    """

    def __init__(self, n, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.n = n
        if not isinstance(n, int):
            raise TypeError(f"Expected an integer value for `n`, got {type(n)}.")

    def call(self, inputs: Tensor, *_args, **_kwargs) -> Tensor:
        input_shape = ibackend.shape(inputs)
        output_shape = [-1, self.n] + input_shape[1:]
        return tf.reshape(inputs, output_shape)


class MultiplyWithLinspace(klayers.Layer):
    def __init__(self, start, end, *args, n=1, axis=-1, **kwargs):
        super().__init__(*args, **kwargs)
        self._start = start
        self._end = end
        self._n = n
        self._axis = axis

    def call(self, inputs: Tensor, *_args, **_kwargs) -> Tensor:
        linspace = self._start + (self._end - self._start) * (
            kbackend.arange(self._n, dtype=kbackend.floatx()) / self._n
        )

        # Make broadcastable.
        shape = np.ones(len(kbackend.int_shape(inputs)))
        shape[self._axis] = self._n
        linspace = kbackend.reshape(linspace, shape)
        return inputs * linspace


class AddGaussianNoise(klayers.Layer):
    "Add Gaussian noise to Tensor. Also applies to test phase."

    def __init__(self, *args, mean: float = 0.0, stddev: float = 1.0, **kwargs):
        super().__init__(*args, **kwargs)
        self.mean = mean
        self.stddev = stddev

    def call(self, inputs: Tensor, *_args, seed=None, **_kwargs) -> Tensor:
        noise = tf.random.normal(
            shape=tf.shape(inputs),
            mean=self.mean,
            stddev=self.stddev,
            dtype=inputs.dtype,
            seed=seed,
        )
        return tf.add(inputs, noise)


class ExtractConv2DPatches(klayers.Layer):
    def __init__(self, kernel_shape, depth, strides, rates, padding, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._kernel_shape = kernel_shape
        self._depth = depth
        self._strides = strides
        self._rates = rates
        self._padding = padding

    def call(self, inputs, *_args, **_kwargs):
        return ibackend.extract_conv2d_patches(
            inputs, self._kernel_shape, self._strides, self._rates, self._padding
        )


class RunningMeans(klayers.Layer):
    """Layer used to keep track of a running mean."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.stateful = True

    def build(self, input_shape: Sequence[ShapeTuple]) -> None:
        means_shape, counts_shape = input_shape

        self.means = self.add_weight(
            shape=means_shape, initializer="zeros", name="means", trainable=False
        )
        self.counts = self.add_weight(
            shape=counts_shape, initializer="zeros", name="counts", trainable=False
        )
        self.built = True

    def call(self, inputs: list[Tensor], *_args, **_kwargs) -> list[Tensor]:
        if len(inputs) != 2:
            raise ValueError(
                "A `RunningMeans` layer should be called on exactly 2 inputs"
            )
        means, counts = inputs
        new_counts = counts + self.counts

        # If new_means are not used for the model output,
        # the following part of the code will be executed after
        # self.counts is updated, therefore we cannot use it
        # hereafter.
        factor_new = ibackend.safe_divide(counts, new_counts, factor=1)
        factor_old = kbackend.ones_like(factor_new) - factor_new
        new_means = self.means * factor_old + means * factor_new

        # Update state.
        self.add_update(
            [
                kbackend.update(self.means, new_means),
                kbackend.update(self.counts, new_counts),
            ]
        )

        return [new_means, new_counts]


class Broadcast(klayers.Layer):
    def call(self, inputs: list[Tensor], *_args, **_kwargs) -> Tensor:
        if len(inputs) != 2:
            raise ValueError("A `Broadcast` layer should be called on exactly 2 inputs")
        target_shapped, x = inputs
        return target_shapped * 0 + x


class MaxNeuronSelection(klayers.Layer):
    """Applied to the last layer of a model, this reduces the output
    to the max neuron activation."""

    def call(self, inputs: Tensor, *_args, **_kwargs) -> Tensor:
        return tf.math.reduce_max(
            inputs, axis=-1, keepdims=False
        )  # max along batch axis


class MaxNeuronIndex(klayers.Layer):
    """Applied to the last layer of a model, this reduces the output
    to the index of the max activated neuron."""

    def call(self, inputs: Tensor, *_args, **_kwargs) -> Tensor:
        return tf.math.argmax(inputs, axis=-1)  # max along batch axis


class NeuronSelection(klayers.Layer):
    """Applied to the last layer of a model, this selects output neurons at given indices
    by wrapping `tf.gather`."""

    def call(self, inputs: list[Tensor], *_args, **_kwargs):
        if len(inputs) != 2:
            raise ValueError(
                "A `NeuronSelection` layer should be called on exactly 2 inputs"
            )
        X, index = inputs
        index_shape = ibackend.shape(index)
        if len(index_shape) != 2 or index_shape[1] != 2:
            raise ValueError(
                "Layer `NeuronSelection` expects index of shape (batch_size, 2),",
                f"got {index_shape}.",
            )
        return tf.gather_nd(X, index)



from innvestigate import analyzer  # noqa
from innvestigate.analyzer import create_analyzer  # noqa
from innvestigate.analyzer.base import NotAnalyzeableModelException  # noqa
from innvestigate.backend.graph import (  # noqa
    model_wo_output_activation,
    model_wo_softmax,
)

__version__ = "2.0.1"



"""Custom types used in iNNvestigate"""

# from tensorflow.python.types import Tensor
from typing import (
    Any,
    Callable,
    List,
    Optional,
    Sequence,
    Tuple,
    TypedDict,
    TypeVar,
    Union,
)

from tensorflow import Tensor
from tensorflow.keras.layers import Layer
from tensorflow.keras.models import Model

T = TypeVar("T")  # Generic type, can be anything

# Keras commonly accepts and returns both tensors and lists of tensors.
# Since iNNvestigate builds on Keras layers, it does too.
# These generic type aliases can be used instead of things like
#   Union[Tensor, List[Tensor]]
OptionalList = Union[T, List[T]]
OptionalSequence = Union[T, Sequence[T]]

# Shapes of tensors are described using tuples of ints (and sometimes Nones).
ShapeTuple = Tuple[Optional[int], ...]

# Type for boolean checks on Keras layers
LayerCheck = Callable[[Layer], bool]

# Used for LRP rules
ReverseRule = Tuple[LayerCheck, Any]  # TODO: replace Any with ReverseMappingBase


class ModelCheckDict(TypedDict):
    """ "Adds type hints to model check dicts."""

    check: LayerCheck
    message: str
    check_type: str


class CondReverseMapping(TypedDict):
    """Adds type hints to conditional reverse mapping dicts."""

    condition: LayerCheck
    mapping: Callable  # TODO: specify type
    name: Optional[str]


class ReverseState(TypedDict):
    """Adds type hints to state used in analyzers of type ReverseAnalyzerBase."""

    layer: Layer
    model: Model
    nid: int
    stop_mapping_at_ids: List[int]


class NodeDict(TypedDict):
    """Adds type hints to NodeDicts.

    Contains the following items:
    * `nid`: the node id.
    * `layer`: the layer creating this node.
    * `Xs`: the input tensors (only valid if not in a nested container).
    * `Ys`: the output tensors (only valid if not in a nested container).
    * `Xs_nids`: the ids of the nodes creating the Xs.
    * `Ys_nids`: the ids of nodes using the according output tensor.
    * `Xs_layers`: the layer that created the according input tensor.
    * `Ys_layers`: the layers using the according output tensor.

    """

    nid: Optional[int]
    layer: Layer
    Xs: List[Tensor]
    Ys: List[Tensor]
    Xs_nids: List[Optional[int]]
    Ys_nids: List[Union[List[int], List[None]]]
    Xs_layers: List[Layer]
    Ys_layers: List[List[Layer]]


class ReverseTensorDict(TypedDict):
    """Typically used in a `Dict[Layer, ReverseTensorDict]`, keeping track
    of reversed tensors propagating relevance to the key-layer.

    * `id`:  tuple of the key-layers Node-ID (the Python object ID)
    and its order in the model reversal.
    * `tensors`: correspond to the tensor(s) propagating relevance to the key-layer.
    * `final_tensor`: sum of all `tensors`. Type-annotated as Optional as it is
    only computed at the end of the model reversal.
    """

    nid: Tuple[int, int]
    tensors: List[Tensor]
    final_tensor: Optional[Tensor]



from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np

__all__ = [
    "project",
    "heatmap",
    "graymap",
    "gamma",
    "clip_quantile",
]


def project(X, output_range=(0, 1), absmax=None, input_is_positive_only=False):
    """Projects a tensor into a value range.

    Projects the tensor values into the specified range.

    :param X: A tensor.
    :param output_range: The output value range.
    :param absmax: A tensor specifying the absmax used for normalizing.
      Default the absmax along the first axis.
    :param input_is_positive_only: Is the input value range only positive.
    :return: The tensor with the values project into output range.
    """

    if absmax is None:
        absmax = np.max(np.abs(X), axis=tuple(range(1, len(X.shape))))
    absmax = np.asarray(absmax)

    mask = absmax != 0
    if mask.sum() > 0:
        X[mask] /= absmax[mask]

    if input_is_positive_only is False:
        X = (X + 1) / 2  # [0, 1]
    X = X.clip(0, 1)

    X = output_range[0] + (X * (output_range[1] - output_range[0]))
    return X


def heatmap(
    X, cmap_type="seismic", reduce_op="sum", reduce_axis=-1, alpha_cmap=False, **kwargs
):
    """Creates a heatmap/color map.

    Create a heatmap or colormap out of the input tensor.

    :param X: A image tensor with 4 axes.
    :param cmap_type: The color map to use. Default 'seismic'.
    :param reduce_op: Operation to reduce the color axis.
      Either 'sum' or 'absmax'.
    :param reduce_axis: Axis to reduce.
    :param alpha_cmap: Should the alpha component of the cmap be included.
    :param kwargs: Arguments passed on to :func:`project`
    :return: The tensor as color-map.
    """
    cmap = plt.cm.get_cmap(cmap_type)

    tmp = X
    shape = tmp.shape

    if reduce_op == "sum":
        tmp = tmp.sum(axis=reduce_axis)
    elif reduce_op == "absmax":
        pos_max = tmp.max(axis=reduce_axis)
        neg_max = (-tmp).max(axis=reduce_axis)
        abs_neg_max = -neg_max
        tmp = np.select(
            [pos_max >= abs_neg_max, pos_max < abs_neg_max], [pos_max, neg_max]
        )
    else:
        raise NotImplementedError()

    tmp = project(tmp, output_range=(0, 255), **kwargs).astype(np.int64)

    if alpha_cmap:
        tmp = cmap(tmp.flatten()).T
    else:
        tmp = cmap(tmp.flatten())[:, :3].T
    tmp = tmp.T

    shape = list(shape)
    shape[reduce_axis] = 3 + alpha_cmap
    return tmp.reshape(shape).astype(np.float32)


def graymap(X, **kwargs):
    """Same as :func:`heatmap` but uses a gray colormap."""
    return heatmap(X, cmap_type="gray", **kwargs)


def gamma(X, gamma=0.5, minamp=0, maxamp=None):
    """
    Apply gamma correction to an input array X
    while maintaining the relative order of entries,
    also for negative vs positive values in X.
    the fxn firstly determines the max
    amplitude in both positive and negative
    direction and then applies gamma scaling
    to the positive and negative values of the
    array separately, according to the common amplitude.

    :param gamma: the gamma parameter for gamma scaling
    :param minamp: the smallest absolute value to consider.
        if not given assumed to be zero (neutral value for relevance,
        min value for saliency, ...). values above and below
        minamp are treated separately.
    :param maxamp: the largest absolute value to consider relative
        to the neutral value minamp
        if not given determined from the given data.
    """

    # prepare return array
    Y = np.zeros_like(X)

    X = X - minamp  # shift to given/assumed center
    if maxamp is None:
        maxamp = np.abs(X).max()  # infer maxamp if not given
    X = X / maxamp  # scale linearly

    # apply gamma correction for both positive and negative values.
    i_pos = X > 0
    i_neg = np.invert(i_pos)
    Y[i_pos] = X[i_pos] ** gamma
    Y[i_neg] = -((-X[i_neg]) ** gamma)

    # reconstruct original scale and center
    Y *= maxamp
    Y += minamp

    return Y


def clip_quantile(X, quantile=1):
    """Clip the values of X into the given quantile."""
    if not isinstance(quantile, (list, tuple)):
        quantile = (quantile, 100 - quantile)

    low = np.percentile(X, quantile[0])
    high = np.percentile(X, quantile[1])
    X[X < low] = low
    X[X > high] = high

    return X
